@@ -46,10 +46,10 @@ def determine_continuum(sme, segment):
     if segment < 0:
         return sme.cscale
 
-    if "spec" not in sme or "mask" not in sme or "wave" not in sme or "uncs" not in sme:
+    if "spec" not in sme or "wave" not in sme:
         # If there is no observation, we have no continuum scale
         warnings.warn("Missing data for continuum fit")
-        cscale = None
+        cscale = [1]
     elif sme.cscale_flag in ["none", -3]:
         cscale = [1]
     elif sme.cscale_flag in ["fix", -1, -2]:
@@ -57,33 +57,38 @@ def determine_continuum(sme, segment):
         cscale = sme.cscale[segment]
     else:
         # fit a line to the continuum points
-        if sme.cscale_flag in ["constant", 0]:
-            ndeg = 0
-        elif sme.cscale_flag in ["linear", 1]:
-            ndeg = 1
-        elif sme.cscale_flag in ["quadratic", 2]:
-            ndeg = 2
-        else:
-            ndeg = sme.cscale_flag
+        ndeg = sme.cscale_degree
 
         # Extract points in this segment
-        x, y, m, u = sme.wave, sme.spec, sme.mask, sme.uncs
+        x, y = sme.wave, sme.spec
+        if "mask" in sme:
+            m = sme.mask
+        else:
+            m = sme.spec.copy()
+            m[:] = sme.mask_value["line"]
+
+        if "uncs" in sme:
+            u = sme.uncs
+        else:
+            u = sme.spec.copy()
+            u[:] = 1
         x, y, m, u = x[segment], y[segment], m[segment], u[segment]
 
         # Set continuum mask
-        if np.all(m != 2):
+        if np.all(m != sme.mask_values["continuum"]):
             # If no continuum mask has been set
             # Use the effective wavelength ranges of the lines to determine continuum points
             logger.info(
-                "No Continuum mask was set, "
-                "Using effective wavelength range of lines to find continuum instead"
+                "No Continuum mask was set in segment %s, "
+                "Using effective wavelength range of lines to find continuum instead",
+                segment,
             )
             cont = get_continuum_mask(x, y, sme.linelist, mask=m)
             # Save mask for next iteration
-            m[cont == 2] = 2
+            m[cont == 2] = sme.mask_values["continuum"]
             logger.debug("Continuum mask points: %i", np.count_nonzero(cont == 2))
 
-        cont = m == 2
+        cont = m == sme.mask_values["continuum"]
         x = x - x[0]
         x, y, u = x[cont], y[cont], u[cont]
 
@@ -91,7 +96,7 @@ def determine_continuum(sme, segment):
         try:
             func = lambda coef: (np.polyval(coef, x) - y) / u
             c0 = np.polyfit(x, y, deg=ndeg)
-            res = least_squares(func, x0=c0, loss="soft_l1")
+            res = least_squares(func, x0=c0)
             cscale = res.x
         except TypeError:
             warnings.warn("Could not fit continuum, set continuum mask?")
@@ -123,6 +128,11 @@ def get_continuum_mask(wave, synth, linelist, threshold=0.1, mask=None):
         True for points between lines and False for points within lines
     """
 
+    if "depth" not in linelist.columns:
+        raise ValueError(
+            "No depth specified in the linelist, can't auto compute the mask"
+        )
+
     if threshold <= 0:
         threshold = 0.01
 
@@ -183,10 +193,10 @@ def determine_radial_velocity(sme, segment, cscale, x_syn, y_syn):
         or None if no observation is present
     """
 
-    if "spec" not in sme or "mask" not in sme or "wave" not in sme or "uncs" not in sme:
+    if "spec" not in sme or "wave" not in sme:
         # No observation no radial velocity
         warnings.warn("Missing data for radial velocity determination")
-        rvel = None
+        rvel = 0
     elif sme.vrad_flag == "none":
         # vrad_flag says don't determine radial velocity
         rvel = sme.vrad[segment]
@@ -198,7 +208,19 @@ def determine_radial_velocity(sme, segment, cscale, x_syn, y_syn):
     else:
         # Fit radial velocity
         # Extract data
-        x, y, m, u = sme.wave, sme.spec, sme.mask, sme.uncs
+        x, y = sme.wave, sme.spec
+        if "mask" in sme:
+            m = sme.mask
+        else:
+            m = sme.spec.copy()
+            m[:] = sme.mask_value["line"]
+
+        if "uncs" in sme:
+            u = sme.uncs
+        else:
+            u = sme.spec.copy()
+            u[:] = 1
+
         # Only this one segment
         x_obs = x[segment]
         y_obs = y[segment].copy()
@@ -242,35 +264,43 @@ def determine_radial_velocity(sme, segment, cscale, x_syn, y_syn):
                 f"Radial velocity flag {sme.vrad_flag} not recognised, expected one of 'each', 'whole', 'none'"
             )
 
+        mask = mask == sme.mask_values["line"]
+        x_obs = x_obs[mask]
+        y_obs = y_obs[mask]
+        u_obs = u_obs[mask]
         y_tmp = np.interp(x_obs, x_syn, y_syn)
 
-        # Get a first rough estimate from cross correlation
-        # Subtract continuum level of 1, for better correlation
-        corr = correlate(
-            y_obs - np.median(y_obs), y_tmp - np.median(y_tmp), mode="same"
-        )
-        offset = np.argmax(corr)
-
-        x1 = x_obs[offset]
-        x2 = x_obs[len(x_obs) // 2]
-        rvel = c_light * (1 - x2 / x1)
-
         rv_bounds = (-100, 100)
-        rvel = np.clip(rvel, *rv_bounds)
+        if np.all(sme.vrad[segment] == 0):
+            # Get a first rough estimate from cross correlation
+            # Subtract continuum level of 1, for better correlation
+            corr = correlate(
+                y_obs - np.median(y_obs), y_tmp - np.median(y_tmp), mode="same"
+            )
+            offset = np.argmax(corr)
 
-        lines = mask != 0
+            x1 = x_obs[offset]
+            x2 = x_obs[len(x_obs) // 2]
+            rvel = c_light * (1 - x2 / x1)
+
+            rvel = np.clip(rvel, *rv_bounds)
+        else:
+            if sme.vrad_flag == "whole":
+                rvel = sme.vrad[0]
+            else:
+                rvel = sme.vrad[segment]
 
         # Then minimize the least squares for a better fit
         # as cross correlation can only find
         def func(rv):
             rv_factor = np.sqrt((1 - rv / c_light) / (1 + rv / c_light))
-            shifted = interpolator(x_obs[lines] * rv_factor)
+            shifted = interpolator(x_obs * rv_factor)
             # shifted = np.interp(x_obs[lines], x_syn * rv_factor, y_syn)
-            resid = (y_obs[lines] - shifted) / u_obs[lines]
+            resid = (y_obs - shifted) / u_obs
             resid = np.nan_to_num(resid, copy=False)
             return resid
 
-        interpolator = util.safe_interpolation(x_syn, y_syn, None)
+        interpolator = lambda x: np.interp(x, x_syn, y_syn)
         res = least_squares(func, x0=rvel, loss="soft_l1", bounds=rv_bounds)
         rvel = res.x[0]
 
@@ -744,4 +774,10 @@ def match_rv_continuum(sme, segments, x_syn, y_syn):
             f"Did not understand cscale_type, expected one of ('whole', 'mask'), but got {sme.cscale_type}."
         )
 
+    # Keep values from unused segments
+    for seg in range(sme.nseg):
+        if seg not in segments:
+            vrad[seg] = sme.vrad[seg]
+            cscale[seg] = sme.cscale[seg]
+
     return cscale, cscale_unc, vrad, vrad_unc
