@@ -13,25 +13,25 @@ import builtins
 import numpy as np
 from tqdm import tqdm
 from scipy.constants import speed_of_light
-from scipy.optimize import OptimizeWarning, least_squares
+from scipy.optimize import OptimizeWarning, least_squares, curve_fit
 from scipy.optimize._numdiff import approx_derivative
+from scipy.special import erf
+from scipy.stats import gennorm
 
 from . import __file_ending__, broadening
 from .abund import Abund
 from .atmosphere.atmosphere import AtmosphereError
 from .atmosphere.savfile import SavFile
 from .atmosphere.krzfile import KrzFile
-from .atmosphere.interpolation import interp_atmo_grid
 from .config import Config
 from .continuum_and_radial_velocity import match_rv_continuum
-from .integrate_flux import integrate_flux
 from .large_file_storage import setup_lfs
 from .iliffe_vector import Iliffe_vector
-from .nlte import update_nlte_coefficients
 from .sme_synth import SME_DLL
 from .uncertainties import uncertainties
-from .util import safe_interpolation, print_to_log
+from .util import print_to_log
 from .synthesize import Synthesizer
+from .nlte import DirectAccessFile
 
 logger = logging.getLogger(__name__)
 
@@ -144,6 +144,11 @@ class SME_Solver:
         synth = synth[segments]
         synth = synth[mask] if mask is not None else synth
 
+        if sme.telluric is not None:
+            tell = sme.telluric[segments]
+            tell = tell[mask] if mask is not None else tell
+            synth *= tell
+
         # TODO: update based on lineranges
         uncs_linelist = 0
 
@@ -206,7 +211,7 @@ class SME_Solver:
 
         return g
 
-    def __get_bounds(self, atmo_file):
+    def get_bounds(self, sme):
         """
         Create Bounds based on atmosphere grid and general rules
 
@@ -219,8 +224,8 @@ class SME_Solver:
         ----------
         param_names : array(str)
             names of the parameters to vary
-        atmo_file : str
-            filename of the atmosphere grid
+        sme : SME_Structure
+            sme structure to get bounds for
 
         Raises
         ------
@@ -241,6 +246,7 @@ class SME_Solver:
             or "logg" in self.parameter_names
             or "monh" in self.parameter_names
         ):
+            atmo_file = sme.atmo.source
             _, ext = splitext(atmo_file)
             atmo_file = self.lfs_atmo.get(atmo_file)
 
@@ -252,11 +258,11 @@ class SME_Solver:
                 bounds["teff"] = teff
 
                 logg = np.unique(atmo_grid.logg)
-                logg = np.min(logg), np.max(logg) + 1
+                logg = np.min(logg), np.inf
                 bounds["logg"] = logg
 
                 monh = np.unique(atmo_grid.monh)
-                monh = np.min(monh), np.max(monh)
+                monh = np.min(monh), np.inf
                 bounds["monh"] = monh
             elif ext == ".krz":
                 # krz atmospheres are fixed to one parameter set
@@ -273,10 +279,24 @@ class SME_Solver:
         # bounds.update({"abund %s" % el: [-10, 11] for el in abund_elem})
 
         result = np.array([[-np.inf, np.inf]] * self.nparam)
-
+        solar = Abund.solar()
         for i, name in enumerate(self.parameter_names):
             if name[:5].lower() == "abund":
-                result[i] = [-10, 11]
+                element = name[5:].strip().capitalize()
+                if element in sme.nlte.elements:
+                    fname = sme.nlte.grids[element]
+                    fname = self.lfs_nlte.get(fname)
+                    grid = DirectAccessFile(fname)
+                    available = grid["abund"]
+                    xmin, xmax = available.min(), available.max()
+                    xmin += solar[element]
+                    xmax += solar[element]
+                    if xmin == xmax:
+                        xmin -= 1
+                        xmax += 1
+                    result[i] = [xmin, xmax]
+                else:
+                    result[i] = [-10, 11]
             elif name[:8].lower() == "linelist":
                 pass
             else:
@@ -289,7 +309,7 @@ class SME_Solver:
         else:
             return [-np.inf, np.inf]
 
-    def __get_scale(self):
+    def get_scale(self):
         """
         Returns scales for each parameter so that values are on order ~1
 
@@ -313,7 +333,7 @@ class SME_Solver:
         ]
         return scales
 
-    def __get_default_values(self, sme):
+    def get_default_values(self, sme):
         """ Default parameter values for each name """
         d = {"teff": 5778, "logg": 4.4, "monh": 0, "vmac": 1, "vmic": 1}
         d.update({f"{el} abund": v for el, v in Abund.solar()().items()})
@@ -326,14 +346,88 @@ class SME_Solver:
             sme[s] if sme[s] is not None else default(s) for s in self.parameter_names
         ]
 
-    def __update_fitresults(self, sme, result):
-        # Update SME structure
-        popt = result.x
-        sme.pfree = np.atleast_2d(popt)  # 2d for compatibility
-        sme.fitparameters = self.parameter_names
+    def estimate_uncertainties(self, unc, resid, deriv):
+        """
+        Estimate the uncertainties by fitting the cumulative distribution of
+        derivative / uncertainties vs. residual / derivative
+        with the generalized normal distribution and use the 68% percentile
+        as the 1 sigma approximation for a normally distributed variable
 
-        for i, name in enumerate(self.parameter_names):
-            sme[name] = popt[i]
+        Parameters
+        ----------
+        unc : array of shape (n,)
+            uncertainties
+        resid : array of shape (n,)
+            residuals of the least squares fit
+        deriv : array of shape (n, p)
+            derivatives (jacobian) of the least squares fit for each parameter
+
+        Returns
+        -------
+        freep_unc : array of shape (p,)
+            uncertainties for each free paramater, in the same order as self.parameter_names
+        """
+
+        freep_name = self.parameter_names
+        nparameters = len(freep_name)
+        freep_unc = np.zeros(nparameters)
+
+        # Cumulative distribution function of the normal distribution
+        # cdf = lambda x, mu, sig: 0.5 * (1 + erf((x - mu) / (np.sqrt(2) * sig)))
+        # std = lambda mu, sig: sig
+
+        def cdf(x, mu, alpha, beta):
+            """
+            Cumulative distribution function of the generalized normal distribution
+            the factor sqrt(2) is a conversion between generalized and regular normal distribution
+            """
+            return gennorm.cdf(x, beta, loc=mu, scale=alpha * np.sqrt(2))
+
+        def std(mu, alpha, beta):
+            """ 1 sigma (68.27 %) quantile, assuming symmetric distribution """
+            interval = gennorm.interval(0.6827, beta, loc=mu, scale=alpha * np.sqrt(2))
+            sigma = (interval[1] - interval[0]) / 2
+            return sigma
+
+        for i in range(nparameters):
+            pder = deriv[:, i]
+            gradlim = np.median(np.abs(pder))
+            idx = np.abs(pder) > gradlim
+            if np.count_nonzero(idx) <= 5:
+                logger.warning(
+                    "Not enough data points with a suitable derivative to determine the uncertainties of %s",
+                    freep_name[i],
+                )
+                continue
+            # Sort pixels according to the change of the i
+            # parameter needed to match the observations
+            idx_sort = np.argsort(resid[idx] / pder[idx])
+            ch_x = resid[idx][idx_sort] / pder[idx][idx_sort]
+            # Weights of the individual pixels also sorted
+            ch_y = np.abs(pder[idx][idx_sort]) / unc[idx][idx_sort]
+            # Cumulative weights
+            ch_y = np.cumsum(ch_y)
+            # Normalized cumulative weights
+            ch_y /= ch_y[-1]
+
+            # Fit the distribution
+            sopt, _ = curve_fit(cdf, ch_x, ch_y)
+
+            hmed = sopt[0]
+            sigma_estimate = std(*sopt)
+            freep_unc[i] = sigma_estimate
+
+            logger.debug(f"{freep_name[i]}: {hmed}, {sigma_estimate}")
+
+        return freep_unc
+
+    def update_fitresults(self, sme, result):
+        # Update SME structure
+        sme.fitresults.clear()
+
+        popt = result.x
+        sme.fitresults.values = popt
+        sme.fitresults.parameters = self.parameter_names
 
         # Determine the covariance
         # hessian == fisher information matrix
@@ -342,27 +436,24 @@ class SME_Solver:
         sig = np.sqrt(covar.diagonal())
 
         # Update fitresults
-        sme.fitresults.clear()
-        sme.fitresults.covar = covar
-        sme.fitresults.grad = result.grad
-        sme.fitresults.pder = result.jac
-        sme.fitresults.resid = result.fun
+        sme.fitresults.covariance = covar
+        sme.fitresults.gradient = result.grad
+        sme.fitresults.derivative = result.jac
+        sme.fitresults.residuals = result.fun
         sme.fitresults.chisq = (
             result.cost * 2 / (sme.spec.size - len(self.parameter_names))
         )
 
-        sme.fitresults.punc = {}
-        sme.fitresults.punc2 = {}
+        sme.fitresults.fit_uncertainties = [np.nan for _ in self.parameter_names]
         for i in range(len(self.parameter_names)):
             # Errors based on covariance matrix
-            sme.fitresults.punc[self.parameter_names[i]] = sig[i]
-            # Errors based on ad-hoc metric
-            tmp = np.abs(result.fun) / np.clip(
-                np.median(np.abs(result.jac[:, i])), 1e-5, None
-            )
-            sme.fitresults.punc2[self.parameter_names[i]] = np.median(tmp)
+            sme.fitresults.fit_uncertainties[i] = sig[i]
+
+        unc = np.concatenate(sme.uncs[sme.mask_good])
+        sme.fitresults.uncertainties = self.estimate_uncertainties(
+            unc, result.fun, result.jac
+        )
 
-        # punc3 = uncertainties(res.jac, res.fun, uncs, param_names, plot=False)
         return sme
 
     def sanitize_parameter_names(self, sme, param_names):
@@ -418,6 +509,7 @@ class SME_Solver:
         sme : SME_Struct
             same sme structure with fit results in sme.fitresults, and best fit spectrum in sme.smod
         """
+
         assert "wave" in sme, "SME Structure has no wavelength"
         assert "spec" in sme, "SME Structure has no observation"
 
@@ -432,6 +524,11 @@ class SME_Solver:
         # Clean parameter values
         if param_names is None:
             param_names = sme.fitparameters
+        if param_names is None or len(param_names) == 0:
+            logger.warning(
+                "No Fit Parameters have been set. Using ('teff', 'logg', 'monh') instead."
+            )
+            param_names = ("teff", "logg", "monh")
         self.parameter_names = self.sanitize_parameter_names(sme, param_names)
 
         self.update_linelist = False
@@ -441,10 +538,15 @@ class SME_Solver:
                 break
 
         # Create appropiate bounds
-        bounds = self.__get_bounds(sme.atmo.source)
-        scales = self.__get_scale()
+        bounds = self.get_bounds(sme)
+        scales = self.get_scale()
         # Starting values
-        p0 = self.__get_default_values(sme)
+        p0 = self.get_default_values(sme)
+        if np.any((p0 < bounds[0]) | (p0 > bounds[1])):
+            logger.warning(
+                "Initial values are incompatible with the bounds, clipping initial values"
+            )
+            p0 = np.clip(p0, bounds[0], bounds[1])
 
         # Get constant data from sme structure
         sme.mask[segments][sme.uncs[segments] == 0] = 0
@@ -457,6 +559,8 @@ class SME_Solver:
         uncs /= spec
 
         logger.info("Fitting Spectrum with Parameters: %s", ",".join(param_names))
+        logger.debug("Initial values: %s", p0)
+        logger.debug("Bounds: %s", bounds)
 
         if (
             sme.wran.min() * (1 - 100 / 3e5) > sme.linelist.wlcent.min()
@@ -489,6 +593,7 @@ class SME_Solver:
                     loss="soft_l1",
                     method="trf",
                     verbose=2,
+                    max_nfev=sme.fitresults.maxiter,
                     args=(sme, spec, uncs, mask),
                     kwargs={"bounds": bounds, "segments": segments},
                 )
@@ -496,14 +601,15 @@ class SME_Solver:
             self.progressbar_jacobian.close()
             # The returned jacobian is "scaled for robust loss function"
             res.jac = self._last_jac
-            sme = self.__update_fitresults(sme, res)
+            for i, name in enumerate(self.parameter_names):
+                sme[name] = res.x[i]
+            sme = self.update_fitresults(sme, res)
             logger.debug("Reduced chi square: %.3f", sme.fitresults.chisq)
             for name, value, unc in zip(
-                self.parameter_names, res.x, sme.fitresults.punc.values()
+                self.parameter_names, res.x, sme.fitresults.uncertainties
             ):
                 logger.info("%s\t%.5f +- %.5g", name.ljust(10), value, unc)
             logger.info("%s\t%s +- %s", "v_rad".ljust(10), sme.vrad, sme.vrad_unc)
-
         elif len(param_names) > 0:
             # This happens when vrad and/or cscale are given as parameters but nothing else
             # We could try to reuse the already calculated synthetic spectrum (if it already exists)
@@ -519,6 +625,6 @@ class SME_Solver:
         return sme
 
 
-def solve(sme, param_names=("teff", "logg", "monh"), segments="all", filename=None):
+def solve(sme, param_names=None, segments="all", filename=None):
     solver = SME_Solver(filename=filename)
     return solver.solve(sme, param_names, segments)
